import unittest
from typing import (
    Callable,
    List,
    Set,
    FrozenSet,
    Mapping,
    Iterable,
    Dict,
    Any,
    Union,
    Tuple,
    NewType,
    Optional,
    Type,
)

from Aspidites._vendor import (
    match_value,
)


class A:
    pass


class B(A):
    pass


class C(A):
    pass


class Z:
    pass

E = NewType("E", B)
lol = NewType("lol", int)
kek = NewType("kek", lol)
cat = NewType("cat", str)
double_kek = NewType('double_kek', Tuple[kek, kek])
composite_kek = NewType("CKek", Union[cat, double_kek])


def annotated(a: int) -> float:
    pass


def big_annotated(a: Tuple[int, float], b: str, c: E) -> double_kek:
    pass


def wrong_annotated(b: str) -> str:
    pass


def not_annotated(q):
    pass


class PampyTypingTests(unittest.TestCase):

    def test_match_any(self):
        self.assertEqual(match_value(Any, 3), (True, [3]))
        self.assertEqual(match_value(Any, 'ok'), (True, ['ok']))

    def test_match_union(self):
        self.assertEqual(match_value(Union[int, str], 3), (True, [3]))
        self.assertEqual(match_value(Union[int, str], 'ok'), (True, ['ok']))
        self.assertEqual(match_value(Union[int, str, float], 5.25), (True, [5.25]))
        self.assertEqual(match_value(Optional[int], None), (True, [None]))
        self.assertEqual(match_value(Optional[int], 1), (True, [1]))
        self.assertEqual(match_value(Optional[int], 1.0), (False, []))

    def test_match_newtype(self):
        self.assertEqual(match_value(lol, 3), (True, [3]))
        self.assertEqual(match_value(kek, 3), (True, [3]))
        self.assertEqual(match_value(double_kek, (13, 37)), (True, [13, 37]))
        self.assertEqual(match_value(composite_kek, "Barsik"), (True, ["Barsik"]))
        self.assertEqual(match_value(composite_kek, (13, 37)), (True, [13, 37]))

    def test_match_type(self):
        self.assertEqual(match_value(Type[int], int), (True, [int]))
        self.assertEqual(match_value(Type[int], 1), (False, []))
        self.assertEqual(match_value(Type[E], "cat"), (False, []))
        self.assertEqual(match_value(Type[A], "cat"), (False, []))

        self.assertEqual(match_value(Type[lol], int), (True, [int]))
        self.assertEqual(match_value(Type[kek], int), (True, [int]))
        self.assertEqual(match_value(Type[kek], str), (False, []))

        self.assertEqual(match_value(Type[A], A), (True, [A]))
        self.assertEqual(match_value(Type[A], B), (True, [B]))
        self.assertEqual(match_value(Type[B], C), (False, []))
        self.assertEqual(match_value(Type[E], B), (True, [B]))

        self.assertEqual(match_value(Type[A], E), (True, [E]))

        self.assertEqual(match_value(Type[E], C), (False, []))
        self.assertEqual(match_value(Type[C], A), (False, []))

        self.assertEqual(match_value(Type[Any], A), (True, [A]))
        self.assertEqual(match_value(Type[Any], Z), (True, [Z]))
        self.assertEqual(match_value(Type[Any], int), (True, [int]))

    def test_match_callable(self):
        self.assertEqual(match_value(Callable[[int], float], annotated), (True, [annotated]))
        self.assertEqual(
            match_value(Callable[[Tuple[int, float], str, E], double_kek], big_annotated), (True, [big_annotated])
        )
        self.assertEqual(match_value(Callable[[int], float], not_annotated), (False, []))
        self.assertEqual(match_value(Callable[[Any], Any], not_annotated), (True, [not_annotated]))
        self.assertEqual(match_value(Callable[[int], float], wrong_annotated), (False, []))

    def test_match_tuple(self):
        self.assertEqual(
            match_value(Tuple[int, str], (1, "a")),
            (True, [1, "a"])
        )
        self.assertEqual(
            match_value(Tuple[int, str], (1, 1)),
            (False, [])
        )
        self.assertEqual(
            match_value(
                Tuple[
                    Union[int, float],
                    Callable[[int], float],
                    Tuple[str, Type[E]]
                ],
                (1.0, annotated, ("ololo", B))
            ),
            (True, [1.0, annotated, "ololo", B])
        )
        self.assertEqual(
            match_value(
                Tuple[
                    Union[int, float],
                    Callable[[int], float],
                    Tuple[str, Type[E]]
                ],
                (1.0, False, ("ololo", B))
            ),
            (False, [])
        )
        self.assertEqual(
            match_value(
                Tuple[
                    Union[int, float],
                    Callable[[int], float],
                    Tuple[str, Type[E]]
                ],
                ("kek", annotated, ("ololo", B))
            ),
            (False, [])
        )
        self.assertEqual(
            match_value(
                Tuple[
                    Union[int, float],
                    Callable[[int], float],
                    Tuple[str, Type[E]]
                ],
                (1, annotated, ("ololo", B, 1488))
            ),
            (False, [])
        )
        self.assertEqual(
            match_value(
                Tuple[
                    Union[int, float],
                    Callable[[int], float],
                    Tuple[str, Type[A]]
                ],
                (1, annotated, ("ololo", E))
            ),
            (True, [1, annotated, "ololo", E])
        )

    def test_match_mapping(self):
        self.assertEqual(match_value(Dict[str, int], {"a": 1, "b": 2}), (True, [{"a": 1, "b": 2}]))
        self.assertEqual(match_value(Mapping[str, lol], {"a": 1, "b": 2}), (True, [{"a": 1, "b": 2}]))
        self.assertEqual(match_value(Dict[str, lol], {"a": 1, "b": 2}), (True, [{"a": 1, "b": 2}]))
        self.assertEqual(match_value(Dict[str, int], {"a": 1.0, "b": 2.0}), (False, []))
        self.assertEqual(match_value(Dict[str, int], {1: 1, 2: 2}), (False, []))
        self.assertEqual(match_value(Mapping[str, int], {1: 1, 2: 2}), (False, []))
        self.assertEqual(match_value(Dict[Union[str, int], int], {1: 1, 2: 2}), (True, [{1: 1, 2: 2}]))
        self.assertEqual(match_value(Dict[Union[str, lol], int], {1: 1, 2: 2}), (True, [{1: 1, 2: 2}]))
        self.assertEqual(
            match_value(Dict[str, Callable[[int], float]], {"a": annotated, "b": annotated}),
            (True, [{"a": annotated, "b": annotated}])
        )

    def test_match_iterable(self):
        self.assertEqual(match_value(List[int], [1, 2, 3]), (True, [[1, 2, 3]]))
        self.assertEqual(match_value(List[int], range(10)), (False, []))
        self.assertEqual(match_value(Iterable[int], [1, 2, 3]), (True, [[1, 2, 3]]))
        self.assertEqual(match_value(Iterable[int], range(10)), (True, [range(10)]))
        self.assertEqual(match_value(List[lol], [1, 2, 3]), (True, [[1, 2, 3]]))
        self.assertEqual(match_value(List[str], [1, 2, 3]), (False, []))
        self.assertEqual(match_value(Iterable[str], [1, 2, 3]), (False, []))
        a_vals = [B(), C(), B()]
        self.assertEqual(match_value(List[A], a_vals), (True, [a_vals]))
        self.assertEqual(match_value(List[str], ["lol", "kek"]), (True, [["lol", "kek"]]))
        self.assertEqual(match_value(List[Union[str, int]], ["lol", "kek"]), (True, [["lol", "kek"]]))
        self.assertEqual(match_value(List[Union[str, int]], [1, 2, 3]), (True, [[1, 2, 3]]))
        self.assertEqual(match_value(List[int], {1, 2, 3}), (False, []))
        self.assertEqual(match_value(Set[int], {1, 2, 3}), (True, [{1, 2, 3}]))
        self.assertEqual(match_value(FrozenSet[int], frozenset([1, 2, 3])), (True, [frozenset([1, 2, 3])]))



from functools import reduce

from Aspidites._vendor._compat import basestring

from ..interface import RValue
from ..syntax import isnumber, W


class Binary(RValue):
    operations = {
        '+': lambda x, y: x + y,
        '-': lambda x, y: x - y,
        '*': lambda x, y: x * y,
        '^': lambda x, y: x ** y,
    }

    precedence = {
        '+': 0,
        '-': 0,
        '*': 1,
        '^': 2,
    }

    def __init__(self, exprs, glyph, where=None):
        assert glyph in  Binary.operations
        for e in exprs:
            assert isinstance(e, RValue)

        self.where = where
        self.exprs = exprs
        self.glyph = glyph
        self.precedence = Binary.precedence[glyph]

    def eval(self, context):  # @ReservedAssignment
        vals = []
        for expr in self.exprs:
            val = expr.eval(context)
            if not isnumber(val):
                raise ValueError('I can only do math with numbers, not %r.' %
                                 val.__class__.__name__)
            vals.append(val)
        operation = Binary.operations[self.glyph]
        return reduce(operation, vals)

    def __repr__(self):
        s = 'Binary(%r,%r)' % (self.exprs, self.glyph)
        return s

    def __str__(self):
        def convert(x):
            if isinstance(x, Binary) and x.precedence < self.precedence:
                return '(%s)' % x
            else:
                return '%s' % x

        s = self.glyph.join(convert(x) for x in self.exprs)
        return s

    @staticmethod
    def parse_action(s, loc, tokens):
        where = W(s, loc)
        l = list(tokens[0])
        exprs = [l.pop(0)]
        while l:
            glyph = l.pop(0)
            assert isinstance(glyph, basestring)
            expr = l.pop(0)
            assert isinstance(expr, RValue)
            exprs.append(expr)

        # noinspection PyUnboundLocalVariable
        return Binary(exprs, glyph, where=where)


class Unary(RValue):

    operations = {
        '-': lambda x: (-x),
    }

    def __init__(self, glyph, expr, where=None):
        assert glyph in Unary.operations
        assert isinstance(expr, RValue)

        self.where = where
        self.expr = expr
        self.glyph = glyph

    def eval(self, context):  # @ReservedAssignment
        val = self.expr.eval(context)
        if not isnumber(val):
            raise ValueError('I can only do math with numbers, not with %r.' %
                   val.__class__.__name__)

        operation = Unary.operations[self.glyph]
        return operation(val)

    def __repr__(self):
        s = 'Unary(%r,%r)' % (self.glyph, self.expr)
        return s

    def __str__(self):
        # XXX: precedence
        return '%s%s' % (self.glyph, self.expr)

    @staticmethod
    def parse_action(s, loc, tokens):
        where = W(s, loc)
        glyph = tokens[0][0]
        assert isinstance(glyph, basestring)
        expr = tokens[0][1]
        assert isinstance(expr, RValue)
        return Unary(glyph, expr, where=where)




import sys
from inspect import ArgSpec

from Aspidites._vendor._compat import basestring

if sys.version_info[0] >= 3:  # pragma: no cover
    from inspect import getfullargspec
    unicode = str

else:  # pragma: no cover
    from collections import namedtuple
    FullArgSpec = namedtuple('FullArgSpec', 'args varargs varkw defaults'
                             ' kwonlyargs kwonlydefaults annotations')
    from inspect import getargspec as _getargspec

    def getargspec(function):
        # print 'hasattr im_func', hasattr(function, 'im_func')
        if hasattr(function, 'im_func'):
            # print('this is a special function : %s' % function)
            # For methods or classmethods drop the first
            # argument from the returned list because
            # python supplies that automatically for us.
            # Note that this differs from what
            # inspect.getargspec() returns for methods.
            # NB: We use im_func so we work with
            #     instancemethod objects also.
            x = _getargspec(function.im_func)
            new_args = x.args[1:]
            spec = ArgSpec(args=new_args, varargs=x.varargs,
                           keywords=x.keywords, defaults=x.defaults)
            return spec

        # print 'calling normal %s' % function
        return _getargspec(function)

    def getfullargspec(function):
        spec = getargspec(function)
        fullspec = FullArgSpec(args=spec.args, varargs=spec.varargs,
                               varkw=spec.keywords,
                               defaults=spec.defaults, kwonlyargs=[],
                               kwonlydefaults=None,
                               annotations={})
        return fullspec


# Backport inspect.getcallargs from Python 2.7 to 2.6
if sys.version_info[:2] == (2, 7):  # pragma: no cover
    # noinspection PyUnresolvedReferences
    from inspect import getcallargs
else:  # pragma: no cover
    inPy3k = sys.version_info[0] == 3

    from inspect import ismethod

    def getcallargs(func, *positional, **named):
        """Get the mapping of arguments to values.

        A dict is returned, with keys the function argument names (including the
        names of the * and ** arguments, if any), and values the respective bound
        values from 'positional' and 'named'.
        """
        args, varargs, varkw, defaults, \
            kwonlyargs, kwonlydefaults, annotations = getfullargspec(func)

        if kwonlyargs:
            raise ValueError("I'm sorry, I don't have the logic to use kwonlyargs. "
                             "Perhapse you can help PyContracts and implement this? Thanks.")

        f_name = func.__name__
        arg2value = {}

        # The following closures are basically because of tuple
        # parameter unpacking.
        assigned_tuple_params = []

        def assign(arg, value):
            if isinstance(arg, basestring):
                arg2value[arg] = value
            else:
                assigned_tuple_params.append(arg)
                value = iter(value)
                for i, subarg in enumerate(arg):
                    try:
                        subvalue = next(value)
                    except StopIteration:
                        raise ValueError('need more than %d %s to unpack' %
                                         (i, 'values' if i > 1 else 'value'))
                    assign(subarg, subvalue)
                try:
                    next(value)
                except StopIteration:
                    pass
                else:
                    raise ValueError('too many values to unpack')

        def is_assigned(arg):
            if isinstance(arg, basestring):
                return arg in arg2value
            return arg in assigned_tuple_params

        if not inPy3k:
            im_self = getattr(func, 'im_self', None)
        else:
            im_self = getattr(func, '__self__', None)

        if ismethod(func) and im_self is not None:
            # implicit 'self' (or 'cls' for classmethods) argument
            positional = (im_self,) + positional
        num_pos = len(positional)
        num_total = num_pos + len(named)
        num_args = len(args)
        num_defaults = len(defaults) if defaults else 0
        for arg, value in zip(args, positional):
            assign(arg, value)
        if varargs:
            if num_pos > num_args:
                assign(varargs, positional[-(num_pos - num_args):])
            else:
                assign(varargs, ())
        elif 0 < num_args < num_pos:
            raise TypeError('%s() takes %s %d %s (%d given)' % (
                f_name, 'at most' if defaults else 'exactly', num_args,
                'arguments' if num_args > 1 else 'argument', num_total))
        elif num_args == 0 and num_total:
            raise TypeError('%s() takes no arguments (%d given)' %
                            (f_name, num_total))
        for arg in args:
            if isinstance(arg, basestring) and arg in named:
                if is_assigned(arg):
                    raise TypeError("%s() got multiple values for keyword "
                                    "argument '%s'" % (f_name, arg))
                else:
                    assign(arg, named.pop(arg))
        if defaults:  # fill in any missing values with the defaults
            for arg, value in zip(args[-num_defaults:], defaults):
                if not is_assigned(arg):
                    assign(arg, value)
        if varkw:
            assign(varkw, named)
        elif named:
            unexpected = next(iter(named))
            if isinstance(unexpected, unicode):
                unexpected = unexpected.encode(sys.getdefaultencoding(), 'replace')
            raise TypeError("%s() got an unexpected keyword argument '%s'" %
                            (f_name, unexpected))
        unassigned = num_args - len([arg for arg in args if is_assigned(arg)])
        if unassigned:
            num_required = num_args - num_defaults
            raise TypeError('%s() takes %s %d %s (%d given)' % (
                f_name, 'at least' if defaults else 'exactly', num_required,
                'arguments' if num_required > 1 else 'argument', num_total))
        return arg2value







# class Extra:
#     loading = False
#
# def load_extra():
#     if not Extra.loading:
#         Extra.loading = True
# #         from . import useful_contracts
#         from .library import miscellaneous_aliases
#         # And after everything else is loaded, load the  utils
#     else:
#         print('already loading...')
    

def parse_contract_string_actual(string):
    from .interface import (Contract, ContractDefinitionError, ContractSyntaxError,
        Where)
    from .main import Storage, _cacheable, check_param_is_string
    from .syntax import ParseException, ParseFatalException, contract_expression

    check_param_is_string(string)

    if string in Storage.string2contract:
        return Storage.string2contract[string]
    try:
        c = contract_expression.parseString(string,
                                            parseAll=True)[0]
        assert isinstance(c, Contract), 'Want Contract, not %r' % c
        if _cacheable(string, c):
            Storage.string2contract[string] = c
        return c
    except ContractDefinitionError as e:
        raise
    except ParseException as e:
        where = Where(string, character=e.loc)
        msg = '%s' % e
        raise ContractSyntaxError(msg, where=where)
    except ParseFatalException as e:
        where = Where(string, character=e.loc)
        msg = '%s' % e
        raise ContractSyntaxError(msg, where=where)





from Aspidites._vendor.RestrictedPython import compile_restricted_eval
from Aspidites._vendor.RestrictedPython import compile_restricted_exec

from Aspidites._vendor.RestrictedPython.Guards import safe_builtins


def _compile(compile_func, source):
    """Compile some source with a compile func."""
    result = compile_func(source)
    assert result.errors == (), result.errors
    assert result.code is not None
    return result.code


def _execute(code, glb, exc_func):
    """Execute compiled code using `exc_func`.

    glb ... globals, gets injected with safe_builtins
    """
    if glb is None:
        glb = {}
    if '__builtins__' not in glb:
        glb['__builtins__'] = safe_builtins.copy()
    if exc_func == 'eval':
        return eval(code, glb)
    else:
        exec(code, glb)
        return glb


def restricted_eval(source, glb=None):
    """Call compile_restricted_eval and actually eval it."""
    code = _compile(compile_restricted_eval, source)
    return _execute(code, glb, 'eval')


def restricted_exec(source, glb=None):
    """Call compile_restricted_eval and actually exec it."""
    code = _compile(compile_restricted_exec, source)
    return _execute(code, glb, 'exec')




# Aspidites
# Copyright (C) 2021 Ross J. Duff

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import contextlib
import os
import typing
from hashlib import sha256
from inspect import isfunction, signature
from os import get_terminal_size
from textwrap import wrap as _wrap

from pyparsing import ParseResults
from pyrsistent import PVector, discard, inc, pmap, pvector, rex

from ._vendor.contracts import new_contract
from ._vendor.fn import F, _
from ._vendor.fn.underscore import ArityError, _Callable
from .templates import _warning


class ContractBreachWarning(RuntimeWarning):
    pass


def wrap(text, width=160, pad=True, padchar=" "):
    """
    Do not remove whitespaces in string but still wrap text to max width.
    Instead of passing the entire text to textwrap.wrap, split and pass each
    line instead. This way list formatting is not mangled by textwrap.wrap.
    """
    wrapped_lines = []
    try:
        for l in text.splitlines():
            line = _wrap(l, width, replace_whitespace=False)
            if pad:
                for s in line:
                    s += padchar * width
            wrapped_lines.extend(line)
    except ArityError:
        wrapped_lines.extend([str(text)])

    return wrapped_lines


def bordered(text, width=160):
    lines = [i for i in wrap(text, width=width)]
    width = max((len(s) for s in lines), default=width) or width
    res = ["╭" + "┉" * width + "╮"]
    for s in lines:
        while len(s) < width:
            s += " "
        res.append("┊" + s + "┊")
    res.append("╰" + "┉" * width + "╯")
    return "\n".join(res)


def format_kwargs(kwargs: "dict", sep=", "):
    return sep + str(kwargs).strip("{} ").replace(":", "=") if len(kwargs) else ""


def format_locals(locals, exc: "Exception"):
    locals_ = dict(filter(lambda x: x[1] != str(exc), locals))
    str_locals = str()
    for k, v_ in locals_.items():
        if str(k).startswith("@"):  # skip @py_assert
            continue
        if isfunction(v_):
            str_locals += k + ": " + str(signature(v_)).replace("'", "") + "\n"
        else:
            str_locals += k + ": " + str(v_) + "\n"
    return str_locals.rstrip("\n")


def create_warning(func, args, kwargs, stack, exc=Exception()):
    _locals = stack[1][0].f_locals.items()
    str_locals = format_locals(_locals, exc)
    func_name = stack[1][0].f_code.co_name
    fname = stack[1][0].f_code.co_filename
    lineno = stack[1][0].f_code.co_firstlineno
    fkwargs = format_kwargs(kwargs)
    if hasattr(func, "__name__"):
        name = func.__name__
    else:
        name = str(func)
    atfault = (
        name
        if isinstance(exc, ArityError)
        else name + "(" + str(args).strip("()") + fkwargs + ")"
    )
    return _warning.safe_substitute(
        file=fname,
        lineno=lineno,
        func=bordered(func_name),
        atfault=bordered(atfault),
        bound=bordered(str_locals),
        tb=bordered(str(exc)),
    )


code = new_contract("code", lambda x: isinstance(x, ParseResults))




from collections import namedtuple
from Aspidites._vendor._compat import IS_CPYTHON
from Aspidites._vendor._compat import IS_PY2
from Aspidites._vendor.RestrictedPython.transformer import RestrictingNodeTransformer

import ast
import warnings


CompileResult = namedtuple(
    'CompileResult', 'code, errors, warnings, used_names')
syntax_error_template = (
    'Line {lineno}: {type}: {msg} at statement: {statement!r}')

NOT_CPYTHON_WARNING = (
    'RestrictedPython is only supported on CPython: use on other Python '
    'implementations may create security issues.'
)


def _compile_restricted_mode(
        source,
        filename='<string>',
        mode="exec",
        flags=0,
        dont_inherit=False,
        policy=RestrictingNodeTransformer):

    if not IS_CPYTHON:
        warnings.warn_explicit(
            NOT_CPYTHON_WARNING, RuntimeWarning, 'RestrictedPython', 0)

    byte_code = None
    collected_errors = []
    collected_warnings = []
    used_names = {}
    if policy is None:
        # Unrestricted Source Checks
        byte_code = compile(source, filename, mode=mode, flags=flags,
                            dont_inherit=dont_inherit)
    elif issubclass(policy, RestrictingNodeTransformer):
        c_ast = None
        allowed_source_types = [str, ast.Module]
        if IS_PY2:  # pragma: no cover
            allowed_source_types.append(unicode)  # NOQA: F821,E501  # PY2 only statement, in Python 2 only module
        if not issubclass(type(source), tuple(allowed_source_types)):
            raise TypeError('Not allowed source type: '
                            '"{0.__class__.__name__}".'.format(source))
        c_ast = None
        # workaround for pypy issue https://bitbucket.org/pypy/pypy/issues/2552
        if isinstance(source, ast.Module):
            c_ast = source
        else:
            try:
                c_ast = ast.parse(source, filename, mode)
            except (TypeError, ValueError) as e:
                collected_errors.append(str(e))
            except SyntaxError as v:
                collected_errors.append(syntax_error_template.format(
                    lineno=v.lineno,
                    type=v.__class__.__name__,
                    msg=v.msg,
                    statement=v.text.strip() if v.text else None
                ))
        if c_ast:
            policy_instance = policy(
                collected_errors, collected_warnings, used_names)
            policy_instance.visit(c_ast)
            if not collected_errors:
                byte_code = compile(c_ast, filename, mode=mode  # ,
                                    # flags=flags,
                                    # dont_inherit=dont_inherit
                                    )
    else:
        raise TypeError('Unallowed policy provided for RestrictedPython')
    return CompileResult(
        byte_code,
        tuple(collected_errors),
        collected_warnings,
        used_names)


def compile_restricted_exec(
        source,
        filename='<string>',
        flags=0,
        dont_inherit=False,
        policy=RestrictingNodeTransformer):
    """Compile restricted for the mode `exec`."""
    return _compile_restricted_mode(
        source,
        filename=filename,
        mode='exec',
        flags=flags,
        dont_inherit=dont_inherit,
        policy=policy)


def compile_restricted_eval(
        source,
        filename='<string>',
        flags=0,
        dont_inherit=False,
        policy=RestrictingNodeTransformer):
    """Compile restricted for the mode `eval`."""
    return _compile_restricted_mode(
        source,
        filename=filename,
        mode='eval',
        flags=flags,
        dont_inherit=dont_inherit,
        policy=policy)


def compile_restricted_single(
        source,
        filename='<string>',
        flags=0,
        dont_inherit=False,
        policy=RestrictingNodeTransformer):
    """Compile restricted for the mode `single`."""
    return _compile_restricted_mode(
        source,
        filename=filename,
        mode='single',
        flags=flags,
        dont_inherit=dont_inherit,
        policy=policy)


def compile_restricted_function(
        p,  # parameters
        body,
        name,
        filename='<string>',
        globalize=None,  # List of globals (e.g. ['here', 'context', ...])
        flags=0,
        dont_inherit=False,
        policy=RestrictingNodeTransformer):
    """Compile a restricted code object for a function.

    Documentation see:
    http://restrictedpython.readthedocs.io/en/latest/usage/index.html#RestrictedPython.compile_restricted_function
    """
    # Parse the parameters and body, then combine them.
    try:
        body_ast = ast.parse(body, '<func code>', 'exec')
    except SyntaxError as v:
        error = syntax_error_template.format(
            lineno=v.lineno,
            type=v.__class__.__name__,
            msg=v.msg,
            statement=v.text.strip() if v.text else None)
        return CompileResult(
            code=None, errors=(error,), warnings=(), used_names=())

    # The compiled code is actually executed inside a function
    # (that is called when the code is called) so reading and assigning to a
    # global variable like this`printed += 'foo'` would throw an
    # UnboundLocalError.
    # We don't want the user to need to understand this.
    if globalize:
        body_ast.body.insert(0, ast.Global(globalize))
    wrapper_ast = ast.parse('def masked_function_name(%s): pass' % p,
                            '<func wrapper>', 'exec')
    # In case the name you chose for your generated function is not a
    # valid python identifier we set it after the fact
    function_ast = wrapper_ast.body[0]
    assert isinstance(function_ast, ast.FunctionDef)
    function_ast.name = name

    wrapper_ast.body[0].body = body_ast.body
    wrapper_ast = ast.fix_missing_locations(wrapper_ast)

    result = _compile_restricted_mode(
        wrapper_ast,
        filename=filename,
        mode='exec',
        flags=flags,
        dont_inherit=dont_inherit,
        policy=policy)

    return result


def compile_restricted(
        source,
        filename='<unknown>',
        mode='exec',
        flags=0,
        dont_inherit=False,
        policy=RestrictingNodeTransformer):
    """Replacement for the built-in compile() function.

    policy ... `ast.NodeTransformer` class defining the restrictions.

    """
    if mode in ['exec', 'eval', 'single', 'function']:
        result = _compile_restricted_mode(
            source,
            filename=filename,
            mode=mode,
            flags=flags,
            dont_inherit=dont_inherit,
            policy=policy)
    else:
        raise TypeError('unknown mode %s', mode)
    for warning in result.warnings:
        warnings.warn(
            warning,
            SyntaxWarning
        )
    if result.errors:
        raise SyntaxError(result.errors)
    return result.code



from Aspidites.tests.helper import restricted_eval


def test_BitAnd():
    assert restricted_eval('5 & 3') == 1


def test_BitOr():
    assert restricted_eval('5 | 3') == 7


def test_BitXor():
    assert restricted_eval('5 ^ 3') == 6


def test_Invert():
    assert restricted_eval('~17') == -18


def test_LShift():
    assert restricted_eval('8 << 2') == 32


def test_RShift():
    assert restricted_eval('8 >> 1') == 4



from ..interface import Contract, ContractNotRespected
from Aspidites._vendor.contracts.syntax import W, S, Keyword, add_contract, add_keyword
from pyparsing import alphanums, Word

__all__ = ['IsInstance', 'isinstance_contract']

class IsInstance(Contract):
    
    """ Checks that one of the superclasses have the specified name. """
     
    def __init__(self, name, where=None):
        Contract.__init__(self, where)
        self.name = name

    def check_contract(self, context, value, silent):
        class_name, bases_names = get_all_super_names(value) 
        
        options = bases_names + [class_name]
        if not self.name in options:
            
            msg = ('Failed check isinstance(%s) for type %r and superclasses %r.' 
                   % (self.name, class_name, bases_names))
            
            # Check it is just a case problem
            lc_name = self.name.lower()
            lc_options = [x.lower() for x in options]
            if lc_name in lc_options:
                msg += '\n***Note: this is just a lower/upper case error!***'

            raise ContractNotRespected(self, msg, value, context)

    def __str__(self):
        return 'isinstance(%s)' % self.name

    def __repr__(self):
        return 'IsInstance(%r)' % self.name

    @staticmethod
    def parse_action(s, loc, tokens):
        where = W(s, loc)
        name = tokens['name']
        return IsInstance(name, where)


def get_all_super_names(value):
    """ Returns name of class, list of names of supers """
    if hasattr(value, '__class__'):
        # old style class
        klass = value.__class__
        class_name = klass.__name__
        bases = get_oldstyle_bases(klass)
        bases_names = [x.__name__ for x in bases]
    else:
        # new style
        t = type(value)
        class_name = t.__name__
        bases_names = [b.__name__ for b in t.mro()]
    return class_name, bases_names

def get_oldstyle_bases(klass):
    todo = [klass]
    res = []
    while todo:
        x = todo.pop(0)
        res.append(x)
        for b in x.__bases__:
            if not b in res:
                todo.append(b)
    return res

Identifier = Word(alphanums + '_')

isinstance_contract = (Keyword('isinstance') - S('(') - Identifier('name') - S(')'))

add_contract(isinstance_contract.setParseAction(IsInstance.parse_action))
add_keyword('isinstance')



import os
import sys
import warnings

import hypothesis

from Aspidites._vendor.contracts import ContractNotRespected
from hypothesis import given, assume, strategies as st
import pytest as pt
from Aspidites.__main__ import get_cy_kwargs, parse_from_dummy, main
import argparse as ap
from pathlib import Path
from Aspidites.parser import parse_module
from Aspidites.templates import lib, setup
from Aspidites.monads import Maybe, Undefined, Surely, SafeMod, SafeDiv, SafeExp
from Aspidites.compiler import compile_module
try:
    from numpy import inf, nan, isinf, isnan
except ImportError:
    from math import inf, nan, isinf, isnan

docker = os.getenv("ASPIDITES_DOCKER_BUILD")

wfile = Path('examples/examples.wom')


@pt.fixture(autouse=True)
def inject_config(request):
    return request.config.rootpath


def setup_code(inject_config):
    if Path(wfile).exists():
        code_ = parse_module(open(wfile, 'r').read())
    else:
        try:
            code_ = parse_module(open(Path("Aspidites/tests") / wfile, 'r').read())
            warnings.warn("Aspidites is being tested in source mode")
        except FileNotFoundError:
            code_ = parse_module(open(Path(inject_config) / wfile, 'r').read())
    return code_


def test_compile_module(inject_config):
    try:
        compile(lib.substitute(code='\n'.join(setup_code(inject_config))), '', 'exec')
    except FileNotFoundError:  # ????
        compile(lib.substitute(code='\n'.join(os.path.join(inject_config(), wfile))), '',
                'exec')


@given(x=st.integers(min_value=-10000,
                     max_value=10000) | st.floats(allow_nan=False),
       y=st.integers(min_value=-10000,
                     max_value=10000) | st.floats(allow_nan=False))
@hypothesis.settings(deadline=None)  # TODO: the specific edge case x=1 y=1 is slow
@pt.mark.filterwarnings('ignore::RuntimeWarning')
def test_safe_math_(x, y):
    assert SafeExp(0, 0) == Undefined()
    assert SafeExp(0, inf) == Undefined()
    assert SafeExp(inf, 0) == Undefined()
    assert SafeMod(x, 0) == Undefined()
    assert SafeDiv(x, 0) == Undefined()
    assert SafeMod(inf, x) == Undefined()
    assume(
        x != 0
        and
        y != 0
    )
    try:
        x ** y
    except OverflowError:  # really big number
        assert SafeExp(x, y) == inf
    else:
        assert SafeExp(x, y) == x ** y

    if isnan(x / y):
        assert SafeDiv(x, y) == Undefined()
    else:
        assert SafeDiv(x, y) == x / y
    assume(
        not isinf(x)
    )
    assert SafeMod(x, y) == x % y
    

@given(x=st.integers())
def test_integer_monad(x):
    assert Undefined() == Undefined()
    assert Undefined() + Undefined() == Undefined()
    assert Undefined() + x == x
    assert Undefined() - Undefined() == Undefined()
    assert Undefined() - x == -x
    assert Undefined() * x == Undefined()
    assert Undefined() * Undefined() == Undefined()
    assert Undefined() / x == Undefined()
    assert Undefined() / Undefined() == Undefined()
    assert Undefined() // x == Undefined()
    assert Undefined().__hash__() == Undefined()
    assert Undefined().__nonzero__() is False
    assert Undefined().__index__() == 0
    assert oct(Undefined()) == oct(0)
    assert complex(Undefined()) == complex()
    assert float(Undefined()) == float()
    assert Surely() == Surely()
    assert Surely() + Surely() == Surely()
    assert Surely() + x == x
    assert Surely() - Surely() == Surely()
    assert Surely() - x == -x
    assert Surely() * x == Surely()
    assert Surely() * Surely() == Surely()
    assert Surely() / x == Surely()
    assert Surely() / Surely() == Surely()
    assert Undefined(x) != Surely(x)
    assert Maybe(x) != x
    assert Surely(x) == Surely(x)
    assert Maybe(x) != Surely(x)
    assert Surely(Maybe(x)) != x
    assert Surely(Undefined()) == Undefined()
    assert Surely(x) == x
    assert -Surely(x) == -x
    assert ~Surely(x) == ~x
    assert ~Maybe(x) != ~x
    assert -Maybe(x) != -x
    assert ~Surely() == Undefined()
    assert oct(Surely(x)) == oct(x)
    assert (not not Surely(x)) == bool(x)
    assert Surely(x) // 1 == x // 1
    assert Surely(x) / 1 == x / 1
    assert Surely(x) * 1 == x * 1
    assert ~Surely(x) == ~x
    assert -Surely(x) == -x
    assert Surely(x) - 1 == x - 1
    assert Surely(x) + 1 == x + 1
    assert (Surely(x) == 1) == (x == 1)
    assert hash(Surely(x)) == hash(x)


def test_cli_help_exit():
    with pt.raises(SystemExit):
        main(['Aspidites', '-h'])


def test_cli_no_arg_exit():
    with pt.raises(SystemExit):
        main(['Aspidites' , ''])


def test_cli_no_target_exit():
    with pt.raises(SystemExit):
        main(['Aspidites' ,'-fpc'])


# @pt.mark.filterwarnings('ignore::RuntimeWarning')
def test_compile_to_shared_object(inject_config):
    pfile_ = Path('examples/compiled.py')
    pfile = pfile_ if Path(wfile).exists() else Path('Aspidites/tests') / pfile_
    kwargs = get_cy_kwargs()
    code = setup_code(inject_config)
    kwargs.update(code=code, fname=pfile, bytecode=True, force=True, c=True, build_requires='', verbose=False)
    try:
        compile_module(**kwargs)
    except FileNotFoundError:
        kwargs.update(code=code, fname=Path(inject_config) / pfile_, bytecode=True)
        compile_module(**kwargs)

    from .examples.compiled import Add, x, y, z, scala, val, div_by_zero, Yield123, Hello, Hello2

    with pt.raises(ContractNotRespected):
        Add(x=6.5, y=12)

    assert [1, 2, 3] == [i for i in Yield123()]
    assert Maybe(Add, 6.5, 12)() == Undefined()
    assert x() == 6
    assert y() == Undefined()
    assert z == 9
    assert Add(x=3, y=2) == 5
    assert val() == Undefined()
    assert div_by_zero() == Undefined()
    Hello()

#
# def teardown_function():
#     for file in glob.glob('examples/compiled.*'):
#         os.remove(file)
#     try:
#         os.remove('examples/setup.py')
#     except FileNotFoundError:
#         pass
